//////////////////////////////////////////////////////////////////////////////
//
// kbwrite
//
//  Simple byte-client -> SRAM writer
//
//  This module stores the byte-scancodes from kbscan.v into the XS40's SRAM.
//  
//  ----------------------------
//  
//  i_byte_rdy initiates a single 'write cycle' :
//
//  1)  Wait for SRAM to become available (i_sram_avail.)
//  2)  assert write-enable
//  3)  assert o_byte_ack (to acknowledge data has been received and accepted)
//  4)  Memory access cycle
//      Wait for XS40 SRAM to be available (monitor the i_sram_avail line)
//      Assert o_sram_wen (1'b0, active-Low), write byte-data to SRAM
//  4)  increment write-address counter, go to 1
//
//  *** In actual board use, the kbwriter is unreliable.  Occasionally it
//      writes data to the *wrong* address.

module kbwrite(
  // inputs
    clk, rstn, 
    i_byte, i_byte_rdy, i_sram_avail,
  // outputs
    o_byte_ack, o_sram_addr, o_sram_wen, o_sram_data
   );

parameter TEXTPAGE_ADDR = 8'h12;  // text page (upper 8 bits of sram addr)

input  clk;                 // system clock
input  rstn;                // async global reset (active low)

input  [7:0]  i_byte;       // Byte-scancode (clk sync'd)
input         i_byte_rdy;   // Keybaord clock line
input         i_sram_avail; // from SRAM, is 'available'

output        o_byte_ack;   // Byte data acknowledgement
output [15:0] o_sram_addr;  // to SRAM, address
output        o_sram_wen;   // to SRAM, write enable (active low)
output [7:0]  o_sram_data;  // to SRAM, data


reg  [7:0] actr;            // address counter, lower-8 bits of srama
wire       actr_inc;        // control, increment actr
wire       actr_reset;      // control, reset actr

wire [15:0] sram_addr;      // SRAM write address

reg        do_write_1d;     // delayed version of do_write

always @ (posedge clk or negedge rstn )
begin
  if ( !rstn )
    actr <= 0;
  else if ( actr_reset )
    actr <= 0;
  else if ( actr_inc )
    actr <= actr + 1;
end
assign actr_reset = 1'b0;
assign actr_inc = do_write_1d;  //wen = active low
  
// assemble complete sram_addr[15:0]   
assign sram_addr[7:0] = actr;
assign sram_addr[15:8] = TEXTPAGE_ADDR;


always @ (posedge clk or negedge rstn )
begin
  if ( !rstn )
    do_write_1d <= 1'b0;
  else
    do_write_1d <= o_byte_ack;
end  // prevent 2 writes in a row

//////////////////////////////////////////////////////////////////////////////
//
// module outputs
//

  assign o_byte_ack = !o_sram_wen;
  assign o_sram_data = i_byte;  // sram write data
  assign o_sram_wen = !(i_sram_avail && i_byte_rdy);  // active low
  assign o_sram_addr = sram_addr;      // write address
endmodule
