/////////////////////////////////////////////////////////////////////////////
//
// mult8x5u.v - Unsigned multiplier 8 x 5 bit
// -------------------
// A multi-cycle General Multiplier - (Shift/add/accumulator implementation)
//
//   This module performs the simple math function "Z = (X * Y);"
// The inputs 'X' and 'Y' are unsigned integers.  The output 'Z' is also
// unsigned, ready 6T later (the status line o_done==1.)  The output 'Z'
// is valid only when o_done is high.
//   The host must hold i_x constant for the entire duration of a 
// multiplication operation!  (This saves on unnecessary registering.)

//   Latency : Measured from i_start to o_done : 5+1 clocks 
//             (in general, one calculation spans <INY_WIDTH>+1 clocks)
//   Thruput : One multiplication cycle per 5+1 clocks
//             (in general, one calculation per <INY_WIDTH>+1 clocks)
// -----
//
//  Xilinx Student Edition 2.1i Synthesis notes :
//
//    (target device is assumed to be XC4000XL-3.)
//
//    (1) The latency can be compressed from 6 clocks to 5 clocks, by
//       directly loading saaZ_reg_next from the SAA-engine to the output Z
//       register.  The penalty is extra combinational logic delay (i.e.
//       decreased clock speed.)
//
//    (2) The multiplier can be extended to compute "Z= (X*Y) + init_val",
//       with minimal extra logic or latency.  Clock-speed will decrease
//       for the same reason cited in (1).
//
//    (3) The base multiplier architecture supports 2's complement signed
//       arithmetic.  However, the necessary modifications aren't discussed 
//       here, because an address calculator does not need 
// -----
//
//   liaor@iname.com - http://members.tripod.com/~liaor  (10/15/2000)
//
//   This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., http://www.gnu.org
//
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// timing waveform
// ---------------
//
//           : 1 : 2 : 3 : 4 : 5 : 6
//        _   _   _   _   _   _   _   _   _   _   _   _   _   _   _   _   _
// clk |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_|
//            ___________________   
// i_x   ----| 'X' <-hold valid->|-------------------
//            ___
// i_y   ----|'Y'|-----------------------------------
//            ___
// i_start __|   |___________________________________
//                                ___
// o_z   ------------------------|'Z'|---------------
//                                ___
// o_done _______________________|   |_______________
//                _______________
// o_busy _______| 2   3   4   5 |___________________

module mult8x5u( 
  // inputs
    clk, rstn, i_x, i_y, i_start,  

  // outputs
    o_z, o_done, o_busy 
    );

// INX_WIDTH & INY_WIDTH are user adjustable.
// ADDR_WIDTH must obey equation : ADDR_WIDTH = log2(INY_WIDTH-1) {round up}

parameter INX_WIDTH  = 8;  // data input X width
parameter INY_WIDTH  = 5;  // data input Y width
parameter ADDR_WIDTH = 3;  // internal counter width
parameter ADD_WIDTH  = (INX_WIDTH+1);  // internal adder output width
parameter ACC_WIDTH  = (INX_WIDTH+INY_WIDTH); // internal accumulator width
parameter OUT_WIDTH  = (ACC_WIDTH); // output width

input                    clk;     // system clock
input                    rstn;    // async reset (active low)
input                    i_start; // start strobe
input  [(INX_WIDTH-1):0] i_x;     // input word 'x'
input  [(INY_WIDTH-1):0] i_y;     // input word 'y'
//input [ (INX_WIDTH-1):0] init_val;

output [(OUT_WIDTH-1):0] o_z;
output                   o_done; // status, 'done' : z is valid this cycle
output                   o_busy; // status, busy line

// SAA engine (shift/add/accumulate) registers and wires
wire [ (INX_WIDTH-1):0] saaX_reg; // input alias (i_x)
reg  [ (ACC_WIDTH-1):0] saaZ_reg;    // shift/add/acc Z register
wire [ (ACC_WIDTH-1):0] saaZ_reg_next;// next state variable
wire                    saa_init_regs; // control, init saaX/Z_reg
wire                    saaZ_shift_and_load; // control, do shift/load
wire                    saa_start;   // control, alias for 'i_start'

// device I/O buses to SAA adder module
wire [ (INX_WIDTH-1):0] adder_ain;  // SAA adder inputs
wire [ (INX_WIDTH-1):0] adder_bin;  // SAA adder inputs
//wire                    adder_cin;  // SAA addsub control (not carryin!)
wire [ (ADD_WIDTH-1):0] adder_out;  // SAA adder output

// SAA state-machine registers/wires
reg                     saa_fsm_ps, saa_fsm_ns; // state machine
reg                     saa_fsm_done, saa_fsm_done_ns;
reg [ (ADDR_WIDTH-1):0] saa_ctr; // bit position counter
wire                    saa_ctr_reset; // control, reset saa_ctr
wire                    saa_ctr_inc;   // control, increment saa_ctr
wire                    saa_ctr_eq_cnt;// status, saa_ctr==(INY_WIDTH-1)

// output registers and wires
//reg  [ (OUT_WIDTH-1):0] z; // output register
//wire                    z_load; // output register load control

// state-machine codes
parameter SM_IDLE = 2'd0;  // IDLE - idle (ready for next input)
parameter SM_SRUN = 2'd1;  // SRUN - busy, running shift/add/acc
//parameter SM_DONE = 2'd3;  // DONE - busy, updating output regs

/////////////////////////////////////////////////////////////////////////////
//
// SAA logic (shift/add/accumulate)
//
// A binary multiplication operation can be decomposed into a series of 
// add & shift operations.  For *positive* terms, this is illustrated below :
//
// -------------------------------------------------------------
// 
//                  X           0010_1011
//                * Y         * 0011_0101
//               -----  ->  --------------     (simplified form)
//                  Z           .........
//
//    The result register, Z, must be wide enough to store the product.
//    This means width(Z) = width(X) + width(Y) - 1;
//
//    1) initialize registers X,Y,Z
//         REG_X       REG_Y            REG_Z
//       0010_1011    0011_0101  0000_0000_0000_0000
//
//    2) RUN saa-engine
//       Let "Yreg0[i]" = 0th bit (LSB) of saaYreg[i]
//       Let "Xreg[i]" = saaXreg ... CONSTANT
//       for( i = 0; i < `DATA_WIDTH'; i = i + 1)
//
//       FIG 1a        0 0 0 0 0 0 0 0   <- accumulator (reg saaZ[i=0])
//                   + 0 0 1 0_1 0 1 1   <- Xreg[] * Yreg0[ i=0 ]
//                  ------------------
//                     0 0 1 0_1 0 1 1   <- saaZ[i=1]
//                 + 0 0 0 0_0 0 0 0     <- Xreg[] * Yreg0[ i=1 ]
//                --------------------
//                   0 0 0 1 0 1 0 1     <- saaZ[i=2]
//               + 0 0 1 0_1 0 1 1       <- Xreg[] * Yreg0[ i=2 ]
//            ------------------------
//                 0 0 1 1 0 1 0 1       <- saaZ[i=3]
//             + 0 0 0 0_0 0 0 0         <- Xreg[] * Yreg0[ i=3 ]
//            ------------------------
//               0 0 0 1 1 0 1 0         <- saaZ[i=4]
//           + 0 0 1 0_1 0 1 1           <- Xreg[] * Yreg0[ i=4 ]
//            ------------------------
//             0 0 1 1 1 0 0 0           <- saaZ[i=5]
//         + 0 0 1 0_1 0 1 1             <- Xreg[] * Yreg0[ i=5 ]
//            ------------------------
//           0 1 0 0 0 1 1 1             <- saaZ[i=6]
//       + 0 0 0 0_0 0 0 0               <- Xreg[] * Yreg0[ i=6 ]
//            ------------------------
//         0 0 1 0 0 0 1 1               <- saaZ[i=7]
//     + 0 0 0 0_0 0 0 0                 <- Xreg[] * Yreg0[ i=7 ]
//            ------------------------
//       0 0 0 1 0 0 0 1                 <- saaZ[i=8]
//
//       Final result = 0 0 0 1 0 0 0 1  1 1 0 0 1 1 1  (binary)
//                      0x8E7 (hex)
//
//      Observations :
//       (a) Only 8-bits are written into reg_Z per iteration.
//       (b) Only 1-bit is read from Xreg per iteration.
//       (c) The Yreg[] is a constant for the entire multiplication process.
//
//      From (a), the saaZ register can be implemented as a shift-register.
//        Each iteration, saaZ shifts right by 1 bit.  The left most bits
//        are loaded with the results of the SAA addition.
//
//      From (b), register X can be 'folded' into the accumulator register.
//        saaZ's lower-half  is initialized with multiplicand X.
//        Tehe shift-right operation of saaZ automatically readies 
//        (Xreg0[ i ]) for the next iteration.
//
//  Additional consideration for negative #s
//  ----------------------------------------
//     The above algorithm works for positive inputs only.  To enable 
//     negative multiplication, some modifications are necessary.  (These
//     modifications aren't necessary for address calculation, so they are
//     not discussed here.) 
//
/////////////////////////////////////////////////////////////////////////////


always @ (posedge clk or negedge rstn )
begin
  if ( !rstn )
    begin
//      saaX_reg <= 0;
//      saaY_reg <= 0;
      saaZ_reg <= 0;
    end
  else if ( saa_init_regs )
    begin
//      saaX_reg <= i_x;

      // init upper-half of saaZ_reg with 'init_val' value
//      saaZ_reg[ (ACC_WIDTH-1):(INY_WIDTH) ] <= init_val;
      saaZ_reg[ (ACC_WIDTH-1):(INY_WIDTH) ] <= 0;
      // init lower-half of saaZ_reg with 'y' value
      saaZ_reg[ (INY_WIDTH-1):0 ] <= i_y;
    end
  else
    begin
      if ( saaZ_shift_and_load )
        saaZ_reg <= saaZ_reg_next;
    end
end
// Right half of saaZ_reg : Perform right shift operation
// Left half of saaZ_reg : Store adder result
assign saaZ_reg_next[ (INY_WIDTH-2):0 ] = saaZ_reg[ (INY_WIDTH-1):1 ];// ASR
assign saaZ_reg_next[ (ACC_WIDTH-1):(INY_WIDTH-1) ] = adder_out; // SUM
assign saaX_reg = i_x;  // i_x must stay constant for duration of mult!

/////////////////////////////////////////////////////////////////////////////
//
// SAA adder unit
//

  assign adder_ain = saaZ_reg[ (ACC_WIDTH-1):(INY_WIDTH) ];
  assign adder_bin = (saaZ_reg[ 0 ] ? saaX_reg : 0);
  assign adder_out = adder_ain + adder_bin;


/////////////////////////////////////////////////////////////////////////////
//
// SAA state machine
//
// The state-machine determines the circuit's progress along the multiply
// operation.  The state-machine relies on a counter (saa_ctr.)

// state-machine NEXT-STATE generator : (combinational logic only)
always @ ( saa_start or saa_fsm_ps or saa_ctr or saa_ctr_eq_cnt )
begin
  saa_fsm_ns = SM_IDLE;
  saa_fsm_done_ns = 1'b0;

  case( saa_fsm_ps ) // synopsys parallel_case full_case

   SM_IDLE : begin   // IDLE state
        if ( saa_start )
          saa_fsm_ns = SM_SRUN;
        else
          saa_fsm_ns = SM_IDLE;
      end

   SM_SRUN : begin   // SAA "run" state
        if ( saa_ctr_eq_cnt )  // positive #s
          begin
            saa_fsm_ns = SM_IDLE; // DONE, back to IDLE!
            saa_fsm_done_ns = 1'b1;
          end
        else
          saa_fsm_ns = SM_SRUN;
      end

//   SM_DONE : begin   // SAA "done" state
//          saa_fsm_ns = SM_IDLE;
//      end

    default :
        saa_fsm_ns = SM_IDLE;
  endcase
end
assign saa_start = i_start;
assign saa_init_regs = ( saa_fsm_ps == SM_IDLE ) && saa_start;
//assign saaY_shift = ( saa_fsm_ps == SM_SRUN );
assign saaZ_shift_and_load = ( saa_fsm_ps == SM_SRUN );

// state-machine output register
always @ (posedge clk or negedge rstn )
begin
  if ( !rstn )
    begin
      saa_fsm_ps <= SM_IDLE;
      saa_fsm_done <= 1'b0;
    end
  else
    begin
      saa_fsm_ps <= saa_fsm_ns;
      saa_fsm_done <= saa_fsm_done_ns;
    end
end

// saa_ctr : track #cycles since start of multiplication
always @ (posedge clk or negedge rstn )
begin
  if ( !rstn )
    saa_ctr <= 0;
  else if ( saa_ctr_reset )
    saa_ctr <= 0;
  else if ( saa_ctr_inc )
    saa_ctr <= saa_ctr + 1;
end
assign saa_ctr_eq_cnt = saa_ctr_inc && ( saa_ctr == (INY_WIDTH-1) );
assign saa_ctr_reset = ( saa_fsm_ps == SM_IDLE );
assign saa_ctr_inc   = ( saa_fsm_ps == SM_SRUN );

/////////////////////////////////////////////////////////////////////////////
//
// module outputs
//

  assign o_busy   = ( saa_fsm_ps != SM_IDLE );
  assign o_done   = saa_fsm_done;
//  assign o_z = saaZ_reg[ (ACC_WIDTH-1):(ACC_WIDTH-OUT_WIDTH) ];
  assign o_z = saaZ_reg >> (ACC_WIDTH-OUT_WIDTH);

endmodule

